"""Converts files to Office format or PDF."""

import asyncio
import subprocess
from pathlib import Path
from typing import Literal

from unichunking.utils import logger


async def convert_file(
    input_path: Path,
    extension: Literal["docx", "xlsx", "pptx", "pdf"],
    max_attempts: int = 5,
) -> Path | None:
    """Converts DOC/XLS/PPT and ODT/ODS/ODP to DOCX/XLSX/PPTX or PDF.

    Args:
        input_path: Path to the local file to convert.
        extension: Output format.
        max_attempts: maximun sttemps before failing

    Returns:
        The path to the local converted file.
    """
    attempt = 0

    while attempt < max_attempts:
        try:

            def run_subprocess() -> None:
                subprocess.run(  # noqa: S603
                    [
                        Path("soffice"),
                        "--headless",
                        "--convert-to",
                        extension,
                        input_path,
                        "--outdir",
                        input_path.parent,
                    ],
                    check=True,
                )

            await asyncio.to_thread(run_subprocess)
            return input_path.with_suffix(f".{extension}")

        except subprocess.CalledProcessError:  # noqa: PERF203
            attempt += 1
            logger.error(
                f"Attempt {attempt}: Couldn't convert the file to {extension}",
            )

    logger.error(
        f"Failed to convert the file to {extension} after {max_attempts} attempts",
    )
    return None



"""Subchunk extraction and formating for any file type."""

from collections.abc import Awaitable
from pathlib import Path
from typing import Any, Callable

from llmax import tokens

from unichunking.settings import unisettings
from unichunking.subchunk_extraction import (
    extract_subchunks_docx,
    extract_subchunks_ipynb,
    extract_subchunks_pdf,
    extract_subchunks_pptx,
    extract_subchunks_txt,
    extract_subchunks_xlsx,
)
from unichunking.types import MatrixTable, StatusManager, SubChunk
from unichunking.utils import logger


def _compute_page_num(
    cur_page_num: int,
    cur_page_size: int,
    subchunk_page: int,
    fake_pages_length: int,
) -> int:
    if fake_pages_length:
        return cur_page_num + (cur_page_size > fake_pages_length)
    return subchunk_page


def _split_table(
    table: MatrixTable,
    size_threshold: int,
) -> tuple[MatrixTable, MatrixTable]:
    table_matrix = table.content
    table_title = table.title

    sub_tab_1, sub_tab_2 = [], []

    first_line = table_matrix[0]
    sub_tab_1 = table_matrix[:2]
    split_after_line = 2
    while (
        tokens.count(str(table_matrix[: split_after_line + 1])) < size_threshold
        and split_after_line < len(table_matrix) - 1
    ):
        sub_tab_1 = table_matrix[: split_after_line + 1]
        split_after_line += 1
    sub_tab_2 = [first_line] + table_matrix[split_after_line:]

    return MatrixTable(table_title, sub_tab_1), MatrixTable(table_title, sub_tab_2)


def _format_table(
    table: MatrixTable,
    chunk_size: int,
    k_max: int,
    min_format: int = 2,
) -> list[str]:
    """Splits a table to fit in the k_max format.

    Keeps the columns/lines information on all subtables to avoid information loss.

    Args:
        table: MatrixTable object to split.
        chunk_size: Size of the current chunk without the table, in tokens.
        k_max: Maximum chunk size, in tokens.
        min_format: Minimum number of lines to keep in a subtable.

    Returns:
        A list of strings representing the subtables created.
    """
    if tokens.count(str(table.read())) + chunk_size <= k_max:
        return [table.read()]

    end_of_cur_chunk, leftover = _split_table(table, size_threshold=k_max - chunk_size)
    tables_done = [end_of_cur_chunk.read()]
    tables_temp = [leftover]

    while len(tables_temp) > 0:
        cur_table = tables_temp.pop()
        cur_table_str = cur_table.read()

        if cur_table.abort():
            tables_done.append(cur_table.read_as_str())

        elif (
            max(len(cur_table.content), len(cur_table.content[0])) <= min_format
            or tokens.count(cur_table_str) <= k_max
        ):
            tables_done.append(cur_table.read())

        else:
            sub_tab_1, sub_tab_2 = _split_table(cur_table, size_threshold=k_max)
            tables_done.append(sub_tab_1.read())
            tables_temp.append(sub_tab_2)

    return tables_done


async def _format_subchunks(
    subchunks: list[SubChunk],
    tables: list[MatrixTable],
    images: list[str],
    k_max: int,
    fake_pages_length: int,
    status_manager: Any = None,
) -> list[SubChunk]:
    """Processes the items extracted by filetype-specific functions to return SubChunks.

    Replaces markers to table/chart/image objects by the corresponding objects.
    Imposes k_max format by splitting subchunks of exceeding length.
    Handles artifical pagination for file types without clear page definition.

    Args:
        subchunks: List of subchhunks extracted by filetype-specific function.
        tables: List of table/chart objects extracted by filetype-specific function.
        images: List of image objects extracted by filetype-specific function.
        k_max: Maximum chunk size, in tokens.
        fake_pages_length: Maximum page size, in tokens, for file types without clear page definition.
        status_manager: Optional, special object to manage task progress.

    Returns:
        A list of SubChunk objects.
    """
    if fake_pages_length:
        logger.debug("Adding artificial page numbers.")

    chunks: list[SubChunk] = []
    cur_page_size = 0
    cur_page_num = 0
    subchunks_progress = 0

    for i_subchunk in range(len(subchunks)):
        new_progress = (int((i_subchunk + 1) / len(subchunks) * 100 / 5)) * 5
        if new_progress > subchunks_progress:
            subchunks_progress = new_progress
            await status_manager.update_status(
                progress=subchunks_progress,
                start=status_manager.start,
                end=status_manager.end,
            )

        subchunk_content = subchunks[i_subchunk].content
        subchunk_content_type = subchunks[i_subchunk].content_type

        match subchunk_content_type:
            case "table":
                table = tables[int(subchunk_content)]
                formatted_tables = _format_table(
                    table=table,
                    chunk_size=0,
                    k_max=k_max,
                )
                subchunk_content = formatted_tables.pop()
                for formatted_table in formatted_tables:
                    new_page_num = _compute_page_num(
                        cur_page_num=cur_page_num,
                        cur_page_size=cur_page_size,
                        subchunk_page=subchunks[i_subchunk].page,
                        fake_pages_length=fake_pages_length,
                    )
                    chunks.append(
                        SubChunk(
                            subchunk_id=len(chunks),
                            content=formatted_table,
                            page=new_page_num,
                            position=subchunks[i_subchunk].position,
                            file_name=subchunks[i_subchunk].file_name,
                            content_type="table",
                        ),
                    )
                    if new_page_num > cur_page_num:
                        cur_page_size = 0
                    cur_page_size += tokens.count(formatted_table)
                    cur_page_num = new_page_num

            case "image":
                subchunk_content = images[int(subchunk_content)]

            case _:
                pass

        if subchunk_content.strip():
            new_page_num = _compute_page_num(
                cur_page_num=cur_page_num,
                cur_page_size=cur_page_size,
                subchunk_page=subchunks[i_subchunk].page,
                fake_pages_length=fake_pages_length,
            )
            chunks.append(
                SubChunk(
                    subchunk_id=len(chunks),
                    content=subchunk_content,
                    page=new_page_num,
                    position=subchunks[i_subchunk].position,
                    file_name=subchunks[i_subchunk].file_name,
                    content_type=subchunks[i_subchunk].content_type,
                ),
            )
            if new_page_num > cur_page_num:
                cur_page_size = 0
            cur_page_size += tokens.count(chunks[-1].content)
            cur_page_num = new_page_num

    return chunks


async def extract_subchunks(  # noqa: C901, PLR0912
    path: Path,
    function: Callable[[str], Awaitable[str]],
    k_max: int = unisettings.chunking.DEFAULT_K_MAX,
    status_manager: Any = None,
) -> tuple[list[SubChunk], int, bool]:
    """Global function that extracts chunks from any document.

    Args:
        path: Path to the local file.
        function: Function to describe images
        k_max: Maximum chunk size, in tokens.
        status_manager: Optional, special object to manage task progress.

    Returns:
        A list of SubChunk objects.
    """
    if not status_manager:
        status_manager = StatusManager(task="Subchunk extraction")

    new_path = path
    extension = path.suffix[1:].lower()
    num_pages = 0

    if extension == "pdf":
        formatted_subchunks, num_pages = await extract_subchunks_pdf(
            path,
            status_manager=status_manager,
            function=function,
        )

    else:
        subchunks, tables, images = [], [], []
        fake_pages_length = unisettings.chunking.FAKE_PAGES_LENGTH

        match extension:
            case "docx" | "doc" | "odt":
                (
                    subchunks,
                    tables,
                    images,
                    new_path,
                    success,
                ) = await extract_subchunks_docx(
                    path,
                    extension,
                    function=function,
                )
                extension = "docx"
                fake_pages_length = 0
                if not success:
                    return [], 0, False

            case "xlsx" | "xls" | "ods":
                (
                    subchunks,
                    tables,
                    images,
                    new_path,
                    success,
                ) = await extract_subchunks_xlsx(
                    path,
                    extension,
                )
                if not success:
                    return [], 0, False

            case "pptx" | "ppt" | "odp":
                (
                    subchunks,
                    tables,
                    images,
                    new_path,
                    success,
                ) = await extract_subchunks_pptx(
                    path,
                    extension,
                    function=function,
                )
                fake_pages_length = 0
                if not success:
                    return [], 0, False

            case "txt" | "md" | "csv":
                subchunks, tables, images = extract_subchunks_txt(
                    path,
                    is_csv=extension == "csv",
                )

            case "ipynb":
                subchunks, tables, images = extract_subchunks_ipynb(path)

            case _:
                pass

        if new_path != path:
            new_path.unlink()
            logger.debug(
                "Deleted local converted file.",
            )

        formatted_subchunks = await _format_subchunks(
            subchunks=subchunks,
            tables=tables,
            images=images,
            k_max=k_max,
            fake_pages_length=fake_pages_length,
            status_manager=status_manager,
        )

    subchunks_final: list[SubChunk] = []

    for subchunk in formatted_subchunks:
        ct = subchunk.content
        ct = (
            ct.replace("\uf0b7", " ")
            .replace("\xe0", "à")
            .replace("\xe7", "ç")
            .replace("\xe8", "è")
            .replace("\xf4", "ô")
            .replace("\xe9", "é")
        )
        ct = ct.replace("   ", " ")
        subchunks_final.append(
            SubChunk(
                subchunk_id=subchunk.subchunk_id,
                content=ct,
                page=subchunk.page,
                position=subchunk.position,
                file_name=subchunk.file_name,
                content_type=subchunk.content_type,
            ),
        )

    return subchunks_final, num_pages, True



"""Extract subchunks from DOCX file."""

import asyncio
import base64
import re
from collections.abc import Awaitable
from pathlib import Path
from typing import Any, Callable

from docx import Document
from docx.oxml import parse_xml
from docx.oxml.ns import qn
from docx.table import Table
from docx.text.paragraph import Paragraph

from unichunking.tools import (
    compact_matrix,
    convert_file,
    extract_charts,
)
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)
from unichunking.utils import logger


def _format_toc(toc: list[tuple[str, str]]) -> str:
    text = "Table of content : \n\n\n"

    for number, title in toc:
        try:
            num = int(number)
        except Exception:  # noqa: BLE001, S112
            continue

        text += f"{'#' * num} {title}\n"

    return text


def _extract_toc(para_xml: str) -> tuple[str, str]:
    p = parse_xml(para_xml)
    p_style_val = p.xpath(".//w:pStyle/@w:val")
    if len(p_style_val) > 0:
        title = p_style_val[0]
        numbers = re.findall(r"\d+", title)
        if numbers:
            number = "".join(numbers)
        else:
            return "", ""
    else:
        return "", ""

    w_t_texts = p.xpath(".//w:t/text()")
    w_t_text_combined = ""
    if w_t_texts:
        w_t_text_combined = "".join(w_t_texts)

    return number, w_t_text_combined


def _docx_to_matrixtable(elt: Table) -> MatrixTable:
    matrix_table: list[list[str]] = []
    for row in elt.rows:
        cur_line = [
            cell.text.replace("\n", " ").replace("\xa0", " ") for cell in row.cells
        ]
        matrix_table.append(cur_line)
    return MatrixTable("", matrix_table)


def _compute_position(section: Any) -> ChunkPosition:
    position = ChunkPosition(0, 0, 1, 1)

    if section.page_width:
        if section.left_margin:
            position.x0 = section.left_margin.inches / section.page_width.inches
        if section.right_margin:
            position.x1 = 1 - section.right_margin.inches / section.page_width.inches
    if section.page_height:
        if section.top_margin:
            position.y0 = section.top_margin.inches / section.page_height.inches
        if section.bottom_margin:
            position.y1 = 1 - section.bottom_margin.inches / section.page_height.inches

    return position


def _heading_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    for run in para.runs:
        if "lastRenderedPageBreak" in run.element.xml or (
            "w:br" in run.element.xml and 'type="page"' in run.element.xml
        ):
            cur_page += 1
        text = run.text
        if run.font.bold:
            text = f"**{text}**"
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )

    return subchunks, cur_page


def _checkbox_paragraph(
    para: Paragraph,
    p: Any,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    seq = """<w:default w:val="1"/>"""

    if seq in p.xml:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml or (
                "w:br" in run.element.xml and 'type="page"' in run.element.xml
            ):
                cur_page += 1
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content="- [X]" + para.text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )
    else:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml or (
                "w:br" in run.element.xml and 'type="page"' in run.element.xml
            ):
                cur_page += 1
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content="- [ ]" + para.text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )

    return subchunks, cur_page


def _list_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    for run in para.runs:
        if "lastRenderedPageBreak" in run.element.xml or (
            "w:br" in run.element.xml and 'type="page"' in run.element.xml
        ):
            cur_page += 1
        espaces_indentation = ""
        if para.paragraph_format.left_indent:
            espaces_indentation = " " * 4
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=espaces_indentation + "- " + para.text.strip() + " ",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
        else:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content="* " + para.text + "\n",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
    return subchunks, cur_page


def _toc_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    if (
        para.style is not None
        and para.style.name is not None
        and "TOC" in para.style.name
    ):
        text = para.text
        if text:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=f"Table des matières : {text}",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
    return subchunks, cur_page


def _handle_paragraph(
    para: Paragraph,
    para_xml: str,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
    image_rels: dict[str, str],
) -> tuple[list[SubChunk], int]:
    p = parse_xml(para_xml)

    if para.style and para.style.name and para.style.name.lower().startswith("toc"):
        subchunks, cur_page = _toc_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif para.style and para.style.name and para.style.name.startswith("Heading"):
        subchunks, cur_page = _heading_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif p.xpath(".//w:checkBox") != []:
        subchunks, cur_page = _checkbox_paragraph(
            para=para,
            p=p,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif para.style and para.style.name == "List Paragraph":
        subchunks, cur_page = _list_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    else:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml:
                cur_page += 1

            text = run.text
            if run.font.bold:
                text = f"**{text}**"
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=text.replace("\t", " ") + " ",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )

    for drawing in p.xpath(".//w:drawing//a:blip"):
        rid = drawing.get(qn("r:embed"))
        if rid in image_rels:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=image_rels[rid],
                    page=cur_page,
                    position=position,
                    file_name=filename,
                    content_type="image",
                ),
            )

    return subchunks, cur_page


async def extract_subchunks_docx(  # noqa: C901, PLR0912, PLR0915
    path: Path,
    extension: str,
    function: Callable[[str], Awaitable[str]],
) -> tuple[list[SubChunk], list[MatrixTable], list[str], Path, bool]:
    """Filetype-specific function : extracts subchunks from a DOCX file.

    For DOC & ODT : local file converted to DOCX, subchunks extracted from DOCX.
    Detects any potential page breaks left by renderers to help with further page computing.

    Args:
        path: Path to the local file.
        extension: File extension.
        function: Function to handle images.

    Returns:
        A tuple containing three lists and a Path:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
        - Path to the processed DOCX file, different from the initial path if a conversion occured.
    """
    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    filename = path.name

    if extension != "docx":
        new_path = await convert_file(path, "docx")
        if new_path is not None:
            path = new_path
        else:
            return [], [], [], Path(), False

    doc = Document(str(path))

    toc: list[tuple[str, str]] = []

    img_idx = 0
    images_jobs: list[dict[str, str | int]] = []
    image_rels: dict[str, str] = {}

    async def handle_image(
        base64_image: str,
        r_id: str,
        img_idx: int,
        semaphore: asyncio.Semaphore,
    ) -> None:
        """Make the async call to describe the image."""
        async with semaphore:
            image_description = await function(base64_image)
            image_rels[r_id] = f"{img_idx}"
            images.append(image_description)

    for rel in doc.part.rels.values():
        if "image" in rel.reltype:
            try:
                image_part = rel.target_part
                image_bytes = image_part.blob
                base64_image = base64.b64encode(image_bytes).decode("utf-8")
                images_jobs.append(
                    {
                        "base64": base64_image,
                        "rId": rel.rId,
                        "img_idx": img_idx,
                    },
                )
                img_idx += 1

            except Exception as e:  # noqa: BLE001
                logger.debug(f"Image couldn't be processed : {e}")

    semaphore = asyncio.Semaphore(5)
    tasks = [
        handle_image(
            str(res["base64"]),
            str(res["rId"]),
            int(res["img_idx"]),
            semaphore,
        )
        for res in images_jobs
    ]
    await asyncio.gather(*tasks)

    cur_page = 0

    for section in doc.sections:
        position = _compute_position(section)

        for elt in section.iter_inner_content():
            if isinstance(elt, Paragraph):
                para_xml = elt._p.xml  # type: ignore # noqa: SLF001
                if 'w:name="_Toc' in para_xml:
                    number, title = _extract_toc(
                        para_xml=para_xml,
                    )
                    if number and title:
                        toc.append((number, title))

                if "<c:chart " in para_xml:
                    for plot in extract_charts(
                        file_path=path,
                        element_xml=para_xml,
                        extension="docx",
                    ):
                        subchunks.append(
                            SubChunk(
                                subchunk_id=len(subchunks),
                                content=f"{len(tables)}",
                                page=cur_page,
                                position=position,
                                file_name=filename,
                                content_type="table",
                            ),
                        )
                        tables.append(plot)

                subchunks, cur_page = _handle_paragraph(
                    para=elt,
                    para_xml=para_xml,
                    position=position,
                    subchunks=subchunks,
                    cur_page=cur_page,
                    filename=filename,
                    image_rels=image_rels,
                )

            else:
                table = compact_matrix(table=_docx_to_matrixtable(elt))
                if table.abort():
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=table.read_as_str(),
                            page=cur_page,
                            position=position,
                            file_name=filename,
                        ),
                    )
                else:
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=f"{len(tables)}",
                            page=cur_page,
                            position=position,
                            file_name=filename,
                            content_type="table",
                        ),
                    )
                    tables.append(table)

    toc_text = _format_toc(toc)

    subchunks.append(
        SubChunk(
            subchunk_id=len(subchunks),
            content=toc_text,
            page=1,
            position=ChunkPosition(0, 0, 1, 1),
            file_name=filename,
            content_type="text",
        ),
    )

    subchunks = [s for s in subchunks if s.content.strip()]

    return subchunks, tables, images, path, True



"""Chhunk formating."""

import copy
from collections.abc import Awaitable
from pathlib import Path
from typing import Any, Callable, Literal
from uuid import uuid4

from llmax import tokens

from unichunking.settings import unisettings
from unichunking.tools import clean_chunk
from unichunking.types import Chunk, StatusManager, SubChunk

from .pages import compute_pages, edit_positions_on_page
from .subchunking import extract_subchunks

Action = Literal["describe", "transcribe"]


def split_chunks_with_overlap(
    chunks: list[SubChunk],
    page_id: int,
    k_min: int = unisettings.chunking.DEFAULT_K_MIN,
    k_max: int = unisettings.chunking.DEFAULT_K_MIN
    + unisettings.chunking.DEFAULT_MIN_MAX_GAP,
    overlap: int = unisettings.chunking.DEFAULT_OVERLAP,
) -> list[Chunk]:
    """Splits the text allowing some overlapping.

    - chunks: les différents chunks extraits du document
    - k_min: nombre de tokens minimal par chunk
    - k_max: nombre maximal de tokens par chunk
    - overlap: nombre de caractères pour l'overlap.
    """
    if not (0 < k_min <= k_max) or overlap < 0:
        message = "Please ensure that 0 < K_MIN <= K_MAX and that overlap >= 0"
        raise ValueError(message)

    chunks_final: list[Chunk] = []

    current_chunk = ""
    current_positions = []

    for it, chunk in enumerate(chunks):
        if tokens.count(current_chunk) + tokens.count(
            chunk.content,
        ) < k_max or it + 1 == len(chunks):
            current_chunk += f" {chunk.content}"
            current_positions.append(chunk.position.to_dict())
            if tokens.count(current_chunk) >= k_min or (it + 1 == len(chunks)):
                curr_chunk = Chunk(
                    chunk_num=len(chunks_final) + 1,
                    content=current_chunk,
                    page_num=page_id,
                    positions=copy.deepcopy(current_positions),
                    embedding=None,
                    chunk_uid=uuid4(),
                    num_tokens=tokens.count(current_chunk),
                )
                curr_chunk = clean_chunk(curr_chunk)
                if curr_chunk.content:
                    chunks_final.append(curr_chunk)
                current_positions = [chunk.position.to_dict()]
                current_chunk = current_chunk[-overlap:] if overlap else ""

        else:
            curr_chunk = Chunk(
                chunk_num=len(chunks_final) + 1,
                content=current_chunk,
                page_num=page_id,
                positions=copy.deepcopy(current_positions),
                embedding=None,
                chunk_uid=uuid4(),
                num_tokens=tokens.count(current_chunk),
            )
            curr_chunk = clean_chunk(curr_chunk)
            if curr_chunk.content:
                chunks_final.append(curr_chunk)
            current_positions = [chunk.position.to_dict()]
            current_chunk = (
                current_chunk[-overlap:] + chunk.content if overlap else chunk.content
            )

    return chunks_final


async def default_function(base64_image: str) -> str:  # noqa: ARG001
    """Default function."""
    return ""


async def build_chunked_pages(
    path: Path,
    k_min: int = unisettings.chunking.DEFAULT_K_MIN,
    overlap: int = unisettings.chunking.DEFAULT_OVERLAP,
    status_manager: Any = None,
    function: Callable[[str], Awaitable[str]] = default_function,
) -> tuple[list[list[Chunk]], int, bool]:
    """Returns a list of Pages and Chunks objects, once extracted, merged when necessary, cleaned and embedded."""
    k_max = k_min + unisettings.chunking.DEFAULT_MIN_MAX_GAP

    extension = path.suffix[1:].lower()
    if unisettings.files.CONTENT_TYPES[extension] == "application/docx":
        extension = "docx"

    chunks: list[Chunk] = []
    chunks_pages: list[list[Chunk]] = []

    if not status_manager:
        status_manager = StatusManager(task="Subchunk extraction")
    status_manager.start = 0
    status_manager.end = 40

    chunks_extracted, num_pages_default, success = await extract_subchunks(
        path,
        k_max=k_max,
        status_manager=status_manager,
        function=function,
    )

    if not success:
        return [[]], 0, False

    pages_numbers = [ch.page for ch in chunks_extracted]
    if len(pages_numbers) == 0 and num_pages_default == 0:
        return [[]], 0, True

    num_pages = max(pages_numbers) + 1 if len(pages_numbers) > 0 else num_pages_default

    status_manager.task = "Chunk formating"
    status_manager.start = 40
    status_manager.end = 75 if extension == "docx" else 100
    for page_id in range(num_pages):
        if page_id % int(num_pages / 30 + 1) == 0:
            page_progress = int((page_id + 1) / num_pages * 100)
            await status_manager.update_status(
                progress=page_progress,
                start=status_manager.start,
                end=status_manager.end,
            )

        relevant_chunks = [ch for ch in chunks_extracted if ch.page == page_id]
        res = split_chunks_with_overlap(
            relevant_chunks,
            page_id=page_id + 1,
            k_min=k_min,
            k_max=k_max,
            overlap=overlap,
        )
        chunks.extend(res)
        chunks_pages.append(res)

    if extension == "docx":
        status_manager.task = "Page computing"
        status_manager.start = 75
        status_manager.end = 100

        chunks, num_pages = await compute_pages(
            chunks=chunks,
            status_manager=status_manager,
        )
        chunks_pages: list[list[Chunk]] = [[] for _ in range(num_pages)]
        for chunk in chunks:
            chunks_pages[chunk.page_num - 1].append(chunk)

    return edit_positions_on_page(chunks_pages, extension), num_pages, True
